import decimal
import json

import orjson
import logging
import os

from flask import Flask, Blueprint, render_template, request, current_app

from flask.logging import default_handler
from flask_sqlalchemy import get_debug_queries
from flaskerk import Flaskerk
from flaskerk.view import APIview
from pydantic import BaseModel, ValidationError

from easi_py_common.core.env import EnvHelper
from werkzeug.utils import find_modules, import_string

from easi_py_common.core.error import ServiceException, intercept_service_exception
from easi_py_common.core.redis import redis_store, redis_ro_store
from easi_py_common.core.sql_alchemy import db
from easi_py_common.core.monkey_doc import validate
from easi_py_common.sqs.producer import producer


class ORJSONDecoder:

    def __init__(self, **kwargs):
        self.options = kwargs

    @staticmethod
    def decode(obj):
        return orjson.loads(obj)


class ORJSONEncoder:

    def __init__(self, **kwargs):
        self.options = kwargs

    @classmethod
    def encode(cls, obj):
        return json.dumps(obj, default=cls.default)
        # try:
        #     return orjson.dumps(obj, default=cls.default)
        # except TypeError:
        #     return json.dumps(obj, default=cls.default)

    @classmethod
    def default(cls, obj):
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        raise TypeError


router_notes = {}
menu_map = {}


def wrap_route(self, rule=None, menu=None, **options):
    if rule is not None:
        url_path = rule
        if self.url_prefix:
            url_path = '{0}{1}'.format(self.url_prefix, rule if rule.startswith('/') else '/{0}'.format(rule))

        url_path = url_path.replace("//", "/")
        methods = options.get('methods', None)
        router_notes[url_path] = methods

        methods = options.get('methods', None)
        if methods is None or ('get' in methods or 'GET' in methods):
            if '<' not in url_path and menu:
                menu_data = list(menu)
                if len(menu_data) == 2:
                    menu_key = menu_data[0]
                    menu_name = menu_data[1]

                    menu_map[url_path] = dict(
                        url_prefix=self.url_prefix,
                        menu_name=menu_name,
                        menu_icon='',
                        url_path=url_path,
                        menu_key=menu_key,
                        is_front_router=0,
                        blue_name=self.name if self.name else ''
                    )
    else:
        if menu:
            menu_data = list(menu)
            if len(menu_data) == 3:
                menu_key = menu_data[0]
                menu_name = menu_data[1]
                menu_url = menu_data[2]

                menu_map[menu_key] = dict(
                    url_prefix=self.url_prefix,
                    menu_name=menu_name,
                    menu_icon='',
                    url_path=menu_url,
                    menu_key=menu_key,
                    is_front_router=1,
                    blue_name=self.name if self.name else ''
                )

    def decorator(f):
        if rule is None:
            return f
        endpoint = options.pop("endpoint", f.__name__)
        self.add_url_rule(rule, endpoint, f, **options)
        return f

    return decorator


def custom_api_doc_view(self):
    assert self.config.ui in self.config._support_ui
    mode = request.args.get('mode')
    ui_file = 'redoc.html'
    if mode == 'swagger':
        ui_file = 'swagger.html'
    return render_template(ui_file, spec_url=self.config.filename)


class AppHelper:
    @staticmethod
    def get_menu():
        return dict(**menu_map)


old_bypass = Flaskerk.bypass


def new_bypass(self, func):
    if not hasattr(func, 'tags'):
        return True
    if 'ignore' in func.tags:
        return True

    return old_bypass(self, func)


class AppWrapper:

    def __init__(self, name: str):
        self.app = self._create_app(name)

    def get_app(self) -> Flask:
        return self.app

    def ctx(self):
        return self.app.app_context()

    def _create_app(self, name: str) -> Flask:
        APIview.get = custom_api_doc_view
        Blueprint.route = wrap_route
        Flaskerk.validate = validate
        Flaskerk.bypass = new_bypass

        app = Flask(name)

        app.config.from_object("config.default")
        env = os.environ.get("ENV")
        if env is not None:
            app.config.from_object("config.{}".format(env))
        else:
            app.config.from_object("config.dev")

        app.json_encoder = ORJSONEncoder
        app.json_decoder = ORJSONDecoder

        db.init_app(app)
        redis_store.init_app(app)
        redis_ro_store.init_app(app)

        self._install_pydantic()

        @app.errorhandler(ServiceException)
        @intercept_service_exception
        def intercept_service_error(e):
            raise e

        @app.errorhandler(ValidationError)
        @intercept_service_exception
        def validation_error(e):
            raise e

        @app.after_request
        def after_request(response):
            if not EnvHelper.is_prod() and 'FLASKY_SLOW_DB_QUERY_TIME' in current_app.config:
                for query in get_debug_queries():
                    if query.duration >= current_app.config['FLASKY_SLOW_DB_QUERY_TIME']:
                        current_app.logger.warning(
                            'Slow query: %s\nParameters: %s\nDuration: %fs\nContext: %s\n' %
                            (query.statement, query.parameters, query.duration,
                             query.context))
            return response

        _format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        logging.basicConfig(level=logging.INFO, format=_format)
        logger = logging.getLogger()
        app.logger.removeHandler(default_handler)
        if app.config['ENV'] == 'production':
            logger.setLevel(logging.ERROR)
        return app

    def start_producer(self):
        producer.start(self.app)

    def register_blueprints(self, path: str = None):
        path = path if path else self.app.import_name
        path = path.replace("-", "_")
        for name in find_modules(path, recursive=True, include_packages=False):
            if 'wsgi' in name:
                continue
            module = import_string(name)
            if hasattr(module, "mod") and isinstance(module.mod, Blueprint):
                self.app.register_blueprint(module.mod)

    def register_blueprint(self, mod: Blueprint):
        for name in find_modules(mod.import_name, recursive=True, include_packages=False):
            import_string(name)
        self.app.register_blueprint(mod)

    def auto_import(self, path: str = None):
        path = path if path else self.app.import_name
        path = path.replace("-", "_")
        for name in find_modules(path, recursive=True, include_packages=False):
            import_string(name)

    def print_router(self):
        for path, methods in router_notes.items():
            self.app.logger.info('{}, {}'.format(path, methods))

    def _install_pydantic(self):
        def orjson_dumps(v, *, default):
            return orjson.dumps(v, default=default).decode()

        BaseModel.Config.orm_mode = True
        BaseModel.Config.json_loads = orjson.loads
        BaseModel.Config.json_dumps = orjson_dumps

    def install_doc(self, doc: Flaskerk):
        doc.register(self.app)



from functools import wraps
from pydantic.generics import GenericModel
from typing import Optional, Callable, TypeVar, Any, Union, Iterable, Type, List, Generic, Tuple

from flask import request, jsonify, make_response, Response, current_app
from pydantic import ValidationError
from typing import Type

from pydantic import BaseModel
from werkzeug.datastructures import ImmutableMultiDict

from easi_py_common.core.error import (
    InvalidIterableOfModelsException,
    ManyModelValidationError,
    JsonBodyParsingError,
)

try:
    from flask_restful import original_flask_make_response as make_response
except ImportError:
    pass

InputParams = TypeVar("InputParams")
DataT = TypeVar('DataT')


class RespWap(GenericModel, Generic[DataT]):
    code: int = 0
    data: Optional[DataT] = {}

    @classmethod
    def __concrete_name__(cls: Type[Any], params: Tuple[Type[Any], ...]) -> str:
        return f'{params[0].__name__.title()}Response'


class RespWaps(GenericModel, Generic[DataT]):
    code: int = 0
    data: List[DataT]

    @classmethod
    def __concrete_name__(cls: Type[Any], params: Tuple[Type[Any], ...]) -> str:
        return f'{params[0].__name__.title()}Response'


class Ok(RespWap, Generic[DataT]):
    # data: Optional[DataT]
    def __init__(self, data: Any = None):
        super(Ok, self).__init__(data=data)


class Oks(RespWap, Generic[DataT]):
    data: List[DataT]

    def __init__(self, data: Any = None):
        super(Oks, self).__init__(data=data)


def make_json_response(
        content: Union[BaseModel, Iterable[BaseModel]],
        status_code: int,
        by_alias: bool,
        exclude_none: bool = False,
        many: bool = False,
) -> Response:
    """serializes model, creates JSON response with given status code"""
    if many:
        js = f"[{', '.join([model.json(exclude_none=exclude_none, by_alias=by_alias) for model in content])}]"
    else:
        js = content.json(exclude_none=exclude_none, by_alias=by_alias)
    response = make_response(js, status_code)
    response.mimetype = "application/json"
    return response


def unsupported_media_type_response(request_cont_type: str) -> Response:
    body = {
        "detail": f"Unsupported media type '{request_cont_type}' in request. "
                  "'application/json' is required."
    }
    return make_response(jsonify(body), 415)


def is_iterable_of_models(content: Any) -> bool:
    try:
        return all(isinstance(obj, BaseModel) for obj in content)
    except TypeError:
        return False


def validate_many_models(model: Type[BaseModel], content: Any) -> List[BaseModel]:
    try:
        return [model(**fields) for fields in content]
    except TypeError:
        # iteration through `content` fails
        err = [
            {
                "loc": ["root"],
                "msg": "is not an array of objects",
                "type": "type_error.array",
            }
        ]
        raise ManyModelValidationError(err)
    except ValidationError as ve:
        raise ManyModelValidationError(ve.errors())


def api_json(
        body: Optional[Type[BaseModel]] = None,
        query: Optional[Type[BaseModel]] = None,
        on_success_status: int = 200,
        exclude_none: bool = False,
        request_body_many: bool = False,
        response_by_alias: bool = False,
):
    """
    Decorator for route methods which will validate query and body parameters
    as well as serialize the response (if it derives from pydantic's BaseModel
    class).

    Request parameters are accessible via flask's `request` variable:
        - request.query_params
        - request.body_params

    Or directly as `kwargs`, if you define them in the decorated function.

    `exclude_none` whether to remove None fields from response
    `response_many` whether content of response consists of many objects
        (e. g. List[BaseModel]). Resulting response will be an array of serialized
        models.
    `request_body_many` whether response body contains array of given model
        (request.body_params then contains list of models i. e. List[BaseModel])

    example::

        from flask import request
        from flask_pydantic import validate
        from pydantic import BaseModel

        class Query(BaseModel):
            query: str

        class Body(BaseModel):
            color: str

        class MyModel(BaseModel):
            id: int
            color: str
            description: str

        ...

        @app.route("/")
        @validate(query=Query, body=Body)
        def test_route():
            query = request.query_params.query
            color = request.body_params.query

            return MyModel(...)

        @app.route("/kwargs")
        @validate()
        def test_route_kwargs(query:Query, body:Body):

            return MyModel(...)

    -> that will render JSON response with serialized MyModel instance
    """

    def decorate(func: Callable[[InputParams], Any]) -> Callable[[InputParams], Any]:
        @wraps(func)
        def wrapper(*args, **kwargs):
            q, b, err = None, None, {}
            query_in_kwargs = func.__annotations__.get("query")
            query_model = query_in_kwargs or query
            if query_model:
                query_params = convert_query_params(request.args, query_model)
                try:
                    q = query_model(**query_params)
                except ValidationError as ve:
                    err["query_params"] = ve.errors()
            body_in_kwargs = func.__annotations__.get("body")
            body_model = body_in_kwargs or body
            if body_model:
                body_params = request.get_json()
                if request_body_many:
                    try:
                        b = validate_many_models(body_model, body_params)
                    except ManyModelValidationError as e:
                        err["body_params"] = e.errors()
                else:
                    try:
                        b = body_model(**body_params)
                    except TypeError:
                        content_type = request.headers.get("Content-Type", "").lower()
                        if content_type != "application/json":
                            return unsupported_media_type_response(content_type)
                        else:
                            raise JsonBodyParsingError()
                    except ValidationError as ve:
                        err["body_params"] = ve.errors()
            request.query_params = q
            request.body_params = b
            if query_in_kwargs:
                kwargs["query"] = q
            if body_in_kwargs:
                kwargs["body"] = b

            if err:
                status_code = current_app.config.get(
                    "FLASK_PYDANTIC_VALIDATION_ERROR_STATUS_CODE", 200
                )
                return make_response(jsonify({'code': 100, 'message': "validation error",
                                              "validation_error": err}), status_code)
            res = func(*args, **kwargs)

            if isinstance(res, list) or isinstance(res, tuple):
                if is_iterable_of_models(res):
                    return make_json_response(
                        res,
                        on_success_status,
                        by_alias=response_by_alias,
                        exclude_none=exclude_none,
                        many=True,
                    )
                else:
                    raise InvalidIterableOfModelsException(res)

            if isinstance(res, BaseModel):
                return make_json_response(
                    res,
                    on_success_status,
                    exclude_none=exclude_none,
                    by_alias=response_by_alias,
                )

            if (
                    isinstance(res, tuple)
                    and len(res) == 2
                    and isinstance(res[0], BaseModel)
            ):
                return make_json_response(
                    res[0],
                    res[1],
                    exclude_none=exclude_none,
                    by_alias=response_by_alias,
                )

            return res

        return wrapper

    return decorate


def convert_query_params(
        query_params: ImmutableMultiDict, model: Type[BaseModel]
) -> dict:
    """
    group query parameters into lists if model defines them

    :param query_params: flasks request.args
    :param model: query parameter's model
    :return: resulting parameters
    """
    return {
        **query_params,
        **{
            key: value
            for key, value in query_params.to_dict(flat=False).items()
            if key in model.__fields__ and model.__fields__[key].is_complex()
        },
    }
